@@ -28,9 +28,9 @@
 #include "interact.h"
 #include "errormsg.h"
 #include "util.h"
-#include "process.h"
+#include "asyprocess.h"
 #include "locate.h"
-#include "parser.h"
+#include "asyparser.h"
 
 namespace vm {
 extern bool indebugger;
@@ -41,30 +41,19 @@ namespace camp {
 extern string tab;
 extern string newline;
 
-enum Mode {NOMODE,INPUT,OUTPUT,UPDATE,BINPUT,BOUTPUT,BUPDATE,XINPUT,XINPUTGZ,
-           XOUTPUT,XUPDATE,OPIPE};
+enum Mode {NOMODE,INPUT,OUTPUT,UPDATE,BINPUT,BOUTPUT,BUPDATE,
+  XINPUT,XOUTPUT,XUPDATE,XINPUTGZ,XOUTPUTGZ,OPIPE};
 
 static const string FileModes[]=
 {"none","input","output","output(update)",
  "input(binary)","output(binary)","output(binary,update)",
- "input(xdr)","input(xdrgz)","output(xdr)","output(xdr,update)","output(pipe)"};
+ "input(xdr)","output(xdr)","output(xdr,update)",
+ "input(xdrgz)","output(xdrgz)","output(pipe)"};
 
 extern FILE *pipeout;
 
-inline void openpipeout()
-{
-  int fd=intcast(settings::getSetting<Int>("outpipe"));
-  if(!pipeout && fd >= 0) pipeout=fdopen(fd,"w");
-  if(!pipeout) {
-    cerr << "Cannot open outpipe " << fd << endl;
-    exit(-1);
-  }
-}
-
-inline string locatefile(string name) {
-  string s=settings::locateFile(name,false,"");
-  return s.empty() ? name : s;
-}
+void openpipeout();
+string locatefile(string name);
 
 class file : public gc {
 protected:
@@ -90,16 +79,11 @@ protected:
 
 public:
 
-  bool Standard() {return standard;}
-
+  bool Standard();
   bool enabled() {return !standard || settings::verbose > 1 ||
       interact::interactive || !settings::getSetting<bool>("quiet");}
 
-  void standardEOF() {
-#if defined(HAVE_LIBREADLINE) && defined(HAVE_LIBCURSES)
-    cout << endl;
-#endif
-  }
+  void standardEOF();
 
   template<class T>
   void purgeStandard(T&) {
@@ -117,50 +101,24 @@ public:
     }
   }
 
-  void purgeStandard(string&) {
-    if(cin.eof())
-      standardEOF();
-  }
+  void purgeStandard(string&);
 
-  void dimension(Int Nx=-1, Int Ny=-1, Int Nz=-1) {
-    if(Nx < -2 || Ny < -2 || Nz < -2) {
-      ostringstream buf;
-      buf << "Invalid array dimensions: " << Nx << ", " << Ny << ", " << Nz;
-      reportError(buf);
-    }
-
-    nx=Nx; ny=Ny; nz=Nz;
-  }
+  void dimension(Int Nx=-1, Int Ny=-1, Int Nz=-1);
 
   file(const string& name, bool check=true, Mode type=NOMODE,
-       bool binary=false, bool closed=false) :
-    name(name), check(check), type(type), linemode(false), csvmode(false),
-    wordmode(false), singlereal(false), singleint(true), signedint(true),
-    closed(closed), standard(name.empty()), binary(binary), nullfield(false),
-    whitespace("") {dimension();}
+       bool binary=false, bool closed=false);
+
+  virtual bool isBinary() {return false;}
+
+  virtual bool isXDR() {return false;}
 
   virtual void open() {}
 
-  void Check() {
-    if(error()) {
-      ostringstream buf;
-      buf << "Cannot open file \"" << name << "\"";
-      reportError(buf);
-    }
-  }
+  void Check();
 
-  virtual ~file() {}
+  virtual ~file();
 
-  bool isOpen() {
-    if(closed) {
-      ostringstream buf;
-      buf << "I/O operation attempted on ";
-      if(name != "") buf << "closed file \'" << name << "\'";
-      else buf << "null file";
-      reportError(buf);
-    }
-    return true;
-  }
+  bool isOpen();
 
   string filename() {return name;}
   virtual bool eol() {return false;}
@@ -177,12 +135,7 @@ public:
 
   string FileMode() {return FileModes[type];}
 
-  void unsupported(const char *rw, const char *type) {
-    ostringstream buf;
-    buf << rw << " of type " << type << " not supported in " << FileMode()
-        << " mode";
-    reportError(buf);
-  }
+  void unsupported(const char *rw, const char *type);
 
   void noread(const char *type) {unsupported("Read",type);}
   void nowrite(const char *type) {unsupported("Write",type);}
@@ -198,6 +151,7 @@ public:
   virtual void readwhite(string&) {noread("string");}
 
   virtual void write(bool) {nowrite("bool");}
+  virtual void write(char) {nowrite("char");}
   virtual void write(Int) {nowrite("int");}
   virtual void write(double) {nowrite("real");}
   virtual void write(const pair&) {nowrite("pair");}
@@ -286,7 +240,7 @@ public:
   bool eof() {return pipeout ? feof(pipeout) : true;}
   bool error() {return pipeout ? ferror(pipeout) : true;}
   void clear() {if(pipeout) clearerr(pipeout);}
-  void flush() {if(pipeout) fflush(pipeout);}
+  void flush();
 
   void seek(Int pos, bool begin=true) {
     if(!standard && pipeout) {
@@ -299,9 +253,7 @@ public:
     return pipeout ? ftell(pipeout) : 0;
   }
 
-  void write(const string& val) {
-    fprintf(pipeout,"%s",val.c_str());
-  }
+  void write(const string& val);
 
   void write(bool val) {
     ostringstream s;
@@ -447,10 +399,7 @@ public:
   void write(guide *val) {*fstream << *val;}
   void write(const transform& val) {*fstream << val;}
 
-  void writeline() {
-    *fstream << newline;
-    if(errorstream::interrupt) throw interrupted();
-  }
+  void writeline();
 };
 
 class ofile : public file {
@@ -465,53 +414,22 @@ public:
 
   ~ofile() {close();}
 
-  void open() {
-    if(standard) {
-      if(mode & std::ios::binary)
-        reportError("Cannot open standard output in binary mode");
-      stream=&cout;
-    } else {
-      name=outpath(name);
-      stream=fstream=new std::ofstream(name.c_str(),mode | std::ios::trunc);
-      stream->precision(settings::getSetting<Int>("digits"));
-      index=processData().ofile.add(fstream);
-      Check();
-    }
-  }
+  void open();
 
   bool text() {return true;}
   bool eof() {return stream->eof();}
   bool error() {return stream->fail();}
 
-  void close() {
-    if(!standard && fstream) {
-      fstream->close();
-      closed=true;
-      delete fstream;
-      fstream=NULL;
-      processData().ofile.remove(index);
-    }
-  }
+  void close();
   void clear() {stream->clear();}
-  Int precision(Int p) {
-    return p == 0 ? stream->precision(settings::getSetting<Int>("digits")) :
-      stream->precision(p);
-  }
+  Int precision(Int p);
   void flush() {stream->flush();}
 
-  void seek(Int pos, bool begin=true) {
-    if(!standard && fstream) {
-      clear();
-      fstream->seekp(pos,begin ? std::ios::beg : std::ios::end);
-    }
-  }
+  void seek(Int pos, bool begin=true);
 
-  size_t tell() {
-    if(fstream)
-      return fstream->tellp();
-    else
-      return 0;
-  }
+  size_t tell();
+
+  bool enabled();
 
   void write(bool val) {*stream << (val ? "true " : "false ");}
   void write(Int val) {*stream << val;}
@@ -531,6 +449,9 @@ public:
   ibfile(const string& name, bool check=true, Mode type=BINPUT,
          std::ios::openmode mode=std::ios::in) :
     ifile(name,check,type,mode | std::ios::binary) {}
+
+  bool isBinary() {return true;}
+
   template<class T>
   void iread(T& val) {
     val=T();
@@ -549,9 +470,21 @@ public:
   }
   void Read(char& val) {iread(val);}
   void Read(string& val) {
-    ostringstream buf;
-    buf << fstream->rdbuf();
-    val=buf.str();
+    size_t n=0;
+    if(wordmode)
+      iread(n);
+    else
+      n=SIZE_MAX;
+    val="";
+    string s;
+    for(size_t i=0; i < n; ++i) {
+      char c;
+      Read(c);
+      if(eof() || error())
+        return;
+      s += c;
+    }
+    val=s;
   }
 
   void Read(double& val) {
@@ -565,6 +498,8 @@ public:
   iobfile(const string& name) :
     ibfile(name,true,BUPDATE,std::ios::in | std::ios::out) {}
 
+  bool isBinary() {return true;}
+
   void flush() {if(fstream) fstream->flush();}
 
   template<class T>
@@ -582,7 +517,13 @@ public:
       else iwrite(unsignedIntcast(val));
     }
   }
-  void write(const string& val) {iwrite(val);}
+  void write(const string& val) {
+    size_t n=val.size();
+    if(wordmode)
+      iwrite(n);
+    for(size_t i=0; i < n; ++i)
+      fstream->write((char *) &val[i],1);
+  }
   void write(const pen& val) {iwrite(val);}
   void write(guide *val) {iwrite(val);}
   void write(const transform& val) {iwrite(val);}
@@ -606,6 +547,8 @@ class obfile : public ofile {
 public:
   obfile(const string& name) : ofile(name,BOUTPUT,std::ios::binary) {}
 
+  bool isBinary() {return true;}
+
   template<class T>
   void iwrite(T val) {
     if(fstream) fstream->write((char *) &val,sizeof(T));
@@ -621,7 +564,13 @@ public:
       else iwrite(unsignedIntcast(val));
     }
   }
-  void write(const string& val) {iwrite(val);}
+  void write(const string& val) {
+    size_t n=val.size();
+    if(wordmode)
+      iwrite(n);
+    for(size_t i=0; i < n; ++i)
+      fstream->write((char *) &val[i],1);
+  }
   void write(const pen& val) {iwrite(val);}
   void write(guide *val) {iwrite(val);}
   void write(const transform& val) {iwrite(val);}
@@ -653,6 +602,8 @@ public:
          xdr::xios::open_mode mode=xdr::xios::in) :
     file(name,check,type,true), fstream(NULL), mode(mode) {}
 
+  bool isXDR() override {return true;}
+
   void open() override {
     name=locatefile(inpath(name));
     fstream=new xdr::ixstream(name.c_str(),mode);
@@ -691,6 +642,30 @@ public:
       return 0;
   }
 
+  void Read(char& val) override {
+    xdr::xbyte b;
+    *fstream >> b;
+    val=b;
+  }
+
+  void Read(string& val) override {
+    size_t n=0;
+    if(wordmode)
+      *fstream >> n;
+    else
+      n=SIZE_MAX;
+    val="";
+    string s;
+    for(size_t i=0; i < n; ++i) {
+      char c;
+      Read(c);
+      if(eof() || error())
+        return;
+      s += c;
+    }
+    val=s;
+  }
+
   void Read(Int& val) override {
     if(signedint) {
       if(singleint) {int ival=0; *fstream >> ival; val=ival;}
@@ -724,31 +699,17 @@ public:
 
 class igzxfile : public ixfile {
 protected:
-  std::vector<char> readData;
+  std::vector<uint8_t> readData;
   size_t const readSize;
   gzFile gzfile;
 public:
-  igzxfile(const string& name, bool check=true, Mode type=XINPUT,
-         xdr::xios::open_mode mode=xdr::xios::in, size_t readSize=32768) :
-         ixfile(name,check,type,mode), readSize(readSize){}
+  igzxfile(const string& name, bool check=true,
+           xdr::xios::open_mode mode=xdr::xios::in, size_t readSize=32768) :
+    ixfile(name,check,XINPUTGZ,mode), readSize(readSize) {}
 
   bool error() override {return !gzfile;}
 
-  void open() override {
-    name=locatefile(inpath(name));
-    gzfile=gzopen(name.c_str(),"rb");
-    Check();
-
-    while(!gzeof(gzfile)) {
-      std::vector<char> tmpBuf(readSize);
-      auto filSz = gzread(gzfile,tmpBuf.data(),readSize);
-      std::copy(tmpBuf.begin(),tmpBuf.begin()+filSz,std::back_inserter(readData));
-    }
-    gzclose(gzfile);
-
-    fstream=new xdr::memixstream(readData);
-    index=processData().ixfile.add(fstream);
-  }
+  void open() override;
 
   void close() override {
     closeFile();
@@ -758,15 +719,7 @@ public:
 
 
 protected:
-  void closeFile()
-  {
-    if(fstream) {
-      fstream->close();
-      closed=true;
-      delete fstream;
-      processData().ixfile.remove(index);
-    }
-  }
+  void closeFile();
 };
 
 class ioxfile : public ixfile {
@@ -784,7 +737,15 @@ public:
 
   void flush() override {if(fstream) ioxfstreamRef->flush();}
 
-  void write(Int val) override{
+  void write(const string& val) override {
+    size_t n=val.size();
+    if(wordmode)
+      *ioxfstreamRef << n;
+    for(size_t i=0; i < n; ++i)
+      *ioxfstreamRef << (xdr::xbyte) val[i];
+  }
+
+  void write(Int val) override {
     if(signedint) {
       if(singleint) *ioxfstreamRef << intcast(val);
       else *ioxfstreamRef << val;
@@ -812,17 +773,21 @@ private:
 };
 
 class oxfile : public file {
+protected:
   xdr::oxstream *fstream;
 public:
-  oxfile(const string& name) : file(name,true,XOUTPUT), fstream(NULL) {}
+  oxfile(const string& name, Mode type=XOUTPUT) : file(name,true,type),
+                                                  fstream(NULL) {}
 
-  void open() {
+  bool isXDR() override {return true;}
+
+  void open() override {
     fstream=new xdr::oxstream(outpath(name).c_str(),xdr::xios::trunc);
     index=processData().oxfile.add(fstream);
     Check();
   }
 
-  void close() {
+  void close() override {
     if(fstream) {
       fstream->close();
       closed=true;
@@ -834,26 +799,34 @@ public:
 
   ~oxfile() {close();}
 
-  bool eof() {return fstream ? fstream->eof() : true;}
-  bool error() {return fstream ? fstream->fail() : true;}
-  void clear() {if(fstream) fstream->clear();}
-  void flush() {if(fstream) fstream->flush();}
+  bool eof() override {return fstream ? fstream->eof() : true;}
+  bool error() override {return fstream ? fstream->fail() : true;}
+  void clear() override {if(fstream) fstream->clear();}
+  void flush() override {if(fstream) fstream->flush();}
 
-  void seek(Int pos, bool begin=true) {
+  void seek(Int pos, bool begin=true) override {
     if(!standard && fstream) {
       clear();
       fstream->seek(pos,begin ? xdr::xios::beg : xdr::xios::end);
     }
   }
 
-  size_t tell() {
+  size_t tell() override {
     if(fstream)
       return fstream->tell();
     else
       return 0;
   }
 
-  void write(Int val) {
+  void write(const string& val) override {
+    size_t n=val.size();
+    if(wordmode)
+      *fstream << n;
+    for(size_t i=0; i < n; ++i)
+      *fstream << (xdr::xbyte) val[i];
+  }
+
+  void write(Int val) override {
     if(signedint) {
       if(singleint) *fstream << intcast(val);
       else *fstream << val;
@@ -862,21 +835,40 @@ public:
       else *fstream << unsignedIntcast(val);
     }
   }
-  void write(double val) {
+  void write(double val) override {
     if(singlereal) *fstream << (float) val;
     else *fstream << val;
   }
-  void write(const pair& val) {
+  void write(const pair& val) override {
     write(val.getx());
     write(val.gety());
   }
-  void write(const triple& val) {
+  void write(const triple& val) override {
     write(val.getx());
     write(val.gety());
     write(val.getz());
   }
 };
 
+class ogzxfile : public oxfile {
+  string name;
+  bool destroyed;
+public:
+  xdr::memoxstream memxdrfile;
+
+  ogzxfile(const string& name, bool singleprecision=false) :
+    oxfile(name,XOUTPUTGZ), name(name), destroyed(false),
+    memxdrfile(singleprecision) {}
+
+  void open() override {
+    fstream=&memxdrfile;
+  }
+
+  void close() override;
+
+  ~ogzxfile() override {close();}
+};
+
 #endif
 
 extern ofile Stdout;
