@@ -24,6 +24,7 @@
 #include "v3dtypes.h"
 #undef transform
 #include "v3dheadertypes.h"
+#include "fileio.h"
 
 namespace camp
 {
@@ -132,6 +133,9 @@ public:
 
   void addPixel(triple const& z0, double width) override;
 
+  void write(const string& s) override;
+  void write(double x) override {}
+
   void precision(int digits) override {}
 
 
@@ -160,9 +164,8 @@ protected:
   xdr::oxstream& getXDRFile() override;
 
 private:
-  xdr::memoxstream memxdrfile;
+  camp::ogzxfile memfile;
   string name;
-  bool destroyed;
   void close() override;
 };
 
